#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/delay.h>
#include <avr/pgmspace.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <string.h>

#define CPUCLOCK      11059200
#define TMC8_CK			  (_BV(CS00))
#define TMC8_CK8		  (_BV(CS01))
#define TMC8_CK64     (_BV(CS00)+_BV(CS01))
#define TMC8_CK256		(_BV(CS02))
#define TMC8_CK1024		(_BV(CS00)+_BV(CS02))

#define MAXCOLOR 7
/*
 * table addresses RGB register triples to display memory
 */
const unsigned char rgbtable_P[] PROGMEM= 
	{24,25,26, 21,22,23, 20,18,19,
	  9,10,11, 12,13,14,  8,16,17,
	 15, 7, 6,  5, 4, 3,  0, 2, 1};

/*
 * global variables
 */
volatile char serbuffer[64];  // serial input buffer
volatile char serin,serout;   // serial buffer pointer

static unsigned char bright[27];  // one byte brightness for every LED
static signed char step[27];      // step width for animating the brightness
static unsigned char dest[27];    // destiniation brightness
static char leds[9];
static char mode;

/*
 * copy displaybuffer to display
 */
INTERRUPT(SIG_OVERFLOW0)
{
	static char intensity=0;
	static char pre=0,keys,keysold;
	int led;
	char bit, v;
	led=0;
	v=0;
	for (bit=0x80; bit!=0; bit>>=1)
		if (intensity<bright[led++]) v|=bit;
	DDRA=v;
	
	v=0;
	for (bit=0x80; bit!=0; bit>>=1)
		if (intensity<bright[led++]) v|=bit;
	DDRB=v;
	
	v=0;
	for (bit=0x80; bit!=0; bit>>=1)
		if (intensity<bright[led++]) v|=bit;
	DDRC=v;
	
	v=0;
	for (bit=0x80; bit!=0x10; bit>>=1)
		if (intensity<bright[led++]) v|=bit;
	DDRD=v+2;

	intensity+=0x04;
	
	// at every full cycle animate the brightness
	if (!intensity)
	{
		pre++;
		if (pre>3)
		{
			pre=0;
  		for (led=0; led<27; led++)
  		{
  			if (step[led]>0)
  			{
  				if (bright[led]<dest[led]-step[led])
  				  bright[led]+=step[led];
  				else
  				{
  				  bright[led]=dest[led];
  				  step[led]=0;
  				}
  			}
  			if (step[led]<0)
  			{
  				if (bright[led]>dest[led]-step[led])
  				  bright[led]+=step[led];
  				else
  				{
  				  bright[led]=dest[led];
  				  step[led]=0;
  				}
  			}
  		}
		}
		// check keypad
		
		keys=PIND;
		if ((!(keys & 0x10)) && (keysold & 0x10)) 
		{
		  serbuffer[(int)serin]='A';
		  serin++;
  		serin&=63;
  	}
		if ((!(keys & 0x08)) && (keysold & 0x08)) 
		{
		  serbuffer[(int)serin]='B';
		  serin++;
  		serin&=63;
  	}
		if ((!(keys & 0x04)) && (keysold & 0x04)) 
		{
		  serbuffer[(int)serin]='C';
		  serin++;
  		serin&=63;
  	}
		keysold=keys;
		
	}
}

/*
 * Switch off all LEDs
 */
void ClearDisplay(void)
{ 
	memset(bright, 0, sizeof(bright));
	memset(dest, 0, sizeof(dest));
	memset(step, 0, sizeof(step));
}

/*
 * initialize LED display
 */
void ledInit(void)
{
	PORTA=0;
	PORTB=0;
	PORTC=0;
	PORTD=0x1d;
	DDRA=0;
	DDRB=0;
	DDRC=0;
	DDRD=2;
  ClearDisplay();	
  // initialize timer
  TCCR0 = TMC8_CK8;  // prescaler = 1024
  TIMSK = _BV(TOIE0);
  sei();
}

/*
 * serial receive isr
 */
SIGNAL(SIG_UART_RECV)
{
  // read the character
  serbuffer[(int)serin]=UDR;
  serin++;
  serin&=63;
}
  

/*
 * serial transmit a character
 */
int UartPutChar(char ch)
{
	// insert a LF if there is a CR
	if (ch=='\n') UartPutChar('\r');
	// wait for transmitter empty
  loop_until_bit_is_set(UCSRA, UDRE);
	// transmit character
	UDR=ch;
	return 0;
}

/*
 * read serial character
 */
int UartGetChar(void)
{
	char ch;
	while (serin==serout);
	ch=serbuffer[(int)serout];
	serout++; if (serout>sizeof(serbuffer)) serout=0;
	return ch;
}

/*
 * check if there is a character in the input buffer
 */
int kbhit(void)
{
	return (serin!=serout);
}

#define UART_BAUD_SELECT(baudRate,xtalCpu) ((xtalCpu)/((baudRate)*16)-1)

/*
 * initialize uart
 */
void UartInit(unsigned long rate)
{
  unsigned int sv;

  sv = UART_BAUD_SELECT(rate,CPUCLOCK);
  UBRRH=(unsigned char) (sv >> 8);
  UBRRL=(unsigned char) sv;

  UCSRB= _BV(RXCIE) | _BV(RXEN) | _BV(TXEN);
  /* Set frame format: asynchronous, 8data, no parity, 1stop bit */
  #ifdef URSEL
  UCSRC = (1<<URSEL)|(3<<UCSZ0);
  #else
  UCSRC = (3<<UCSZ0);
  #endif 
  // dummy read to reset error bits
  sv=UDR;
	fdevopen(UartPutChar, UartGetChar, 0);
}

char delay(void)
{
	int t1,t2;
	for (t2=0;t2<5000;t2++)
  	for (t1=0;t1<3;t1++)
  	{
  		if (kbhit()) return 1;
  	}
  return 0;
}

/*
 * set color to rgb
 */
void setled(char led, char r, char g, char b)
{
	int adr=(int)rgbtable_P+3*led;
	int index;
	index=PRG_RDB(adr++);
	dest[index]=bright[index]=r;
	step[index]=0;
	index=PRG_RDB(adr++);
	dest[index]=bright[index]=g;
	step[index]=0;
	index=PRG_RDB(adr++);
	dest[index]=bright[index]=b;
	step[index]=0;
}

/*
 * smoothly alter color to rgb
 */
void dimled(char led, char r, char g, char b)
{
	int adr=(int)rgbtable_P+3*led;
	int index;
	
	index=PRG_RDB(adr++);
  dest[index]=r;
  step[index]=((int)r-bright[index])>>6;

	index=PRG_RDB(adr++);
  dest[index]=g;
  step[index]=((int)g-bright[index])>>6;

	index=PRG_RDB(adr++);
  dest[index]=b;
  step[index]=((int)b-bright[index])>>6;
}

/*
 * check if all animations are ready
 */
char dimready(void)
{
	int led;
	for (led=0; led<27; led++)
		if (step[led]) return 0;
	return 1;
}

void SaveColor(int color, char r, char g, char b)
{
	char *addr=(char *)(3*color);
	eeprom_write_byte(addr++,r);	
	eeprom_write_byte(addr++,g);	
	eeprom_write_byte(addr,b);
}

void LoadColor(int color, char *r, char *g, char *b)
{
	char *addr=(char *)(3*color);
	*r=eeprom_read_byte(addr++);
	*g=eeprom_read_byte(addr++);
	*b=eeprom_read_byte(addr);
}

void ColorEdit(void)
{
	static int color=0;
	char r,g,b,ch;
	ClearDisplay();
	puts_P(PSTR("Select Color: +,-\n"
	            "Control the color: 1=Red-, 4=Red+, 2=Green-, 5=Green+, 3=Blue-, 6=Blue+\n"
	            "Press <ESC> to exit or <S> to save\n"));
	LoadColor(color,&r,&g,&b);
	do
	{
		setled(4,r,g,b);
		printf_P(PSTR("\rColor %3i : R:%03i, G:%03i, B:%03i"),color,r,g,b);
		ch=getchar(); ch=toupper(ch);
		switch (ch)
		{
			case '+' :
			case 'A' : if (color<MAXCOLOR) color++; else color=0; LoadColor(color,&r,&g,&b); break;
			case '-' : if (color>0) color--; else color=MAXCOLOR; LoadColor(color,&r,&g,&b); break;
			case '1' : if (r>3) {if (r==255) r=252; else r-=4;} break;
			case '4' : if (r<252) r+=4; else r=255; break;
			case '2' : if (g>3) {if (g==255) g=252; else g-=4;} break;
			case '5' : if (g<252) g+=4; else g=255; break;
			case '3' : if (b>3) {if (b==255) b=252; else b-=4;} break;
			case '6' : if (b<252) b+=4; else b=255; break;
			case 'S' :
			case 'C' : SaveColor(color,r,g,b); break;
		}
	} while ((ch!=27) && (ch!='S') && (ch!='C'));
	putchar('\n');
}

void Animation(void)
{
	static char cycle=0, wait=0;
	static int color1, color2,color1old, color2old;
	int i;
	char r,g,b;
	color1old=color2old=255;
//  do
//  {
 		if (dimready())
 		{
			if (wait)
			{
				delay();
				wait--;
			}
			else
			{
    		switch (mode)
    		{
    			case 0: // Color Edit: do nothing
    				return;
    			case 1: 
    			case 2: 
    			  switch (cycle)
    			  {
    			  	case 0:
    			  	  // calculate two different colors
            		do
            			color1=random()%(MAXCOLOR+1);
            		while ((color1==color1old) || (color1==color2old));
            		do
            			color2=random()%(MAXCOLOR+1);
            		while ((color2==color1) || (color2==color1old) || (color2==color2old));
            		// generate a two colored random pattern
            		for (i=0; i<9; i++)
            		{
            			if (random()&1)
            				leds[i]=color1;
            			else
            				leds[i]=color2;
            		}
            		// get the real RGB color
            		LoadColor(color1,&r,&g,&b);
            		// and load the first color into the LED array
            		for (i=0; i<9; i++)
            			if (leds[i]==color1) dimled(i,r,g,b);
            		// now wait for the next cycle
            		cycle++;
            		wait=50;
            		break;
            	case 1:
            		// get the real RGB color
            		LoadColor(color2,&r,&g,&b);
            	  // load the second color into the LED array
            		for (i=0; i<9; i++)
            			if (leds[i]==color2) dimled(i,r,g,b);
            		// now wait for the next cycle
            		color1old=color1;
            		color2old=color2;
            		if (mode==1)
            			cycle++;
            		else
            		  cycle=0;
            		wait=50;
            		break;
            	case 2:
            	  // now dim down all LEDs of first color
            		for (i=0; i<9; i++)
            			if (leds[i]==color1) dimled(i,0,0,0);
            		// now wait for the next cycle
            		cycle++;
            		wait=50;
            		break;
            	case 3:
  							// now dim down all LEDs of second color
            		for (i=0; i<9; i++)
            			if (leds[i]==color2) dimled(i,0,0,0);
            		// now wait for the next cycle
            		cycle=0;
            		wait=50;
            		break;
            }
          	break;
          case 3:
            for (i=0; i<9; i++) 
            {
            	leds[i]=random()%(MAXCOLOR+1);
           		LoadColor(leds[i],&r,&g,&b);
            	dimled(i,r,g,b);
            }
            wait=50;
            break;
        }
      }
	  }
// 	} while (1);
}

int main(void)
{
	char ch;
	UartInit(19200);
	ledInit();
	while (kbhit()) getchar();
	puts_P(PSTR("RGB LED Firmware 1.0\n"));
	mode=1;
	while (1)
	{
		printf_P(PSTR("\nA    - Next mode\n"
		                "B    - Edit Color\n"
		                "1..3 - Select mode\n"
		                "C    - Power on/off\n"
		                "Mode %i>"),mode);
		while (!kbhit()) Animation();
		ch=getchar(); ch=toupper(ch);
		switch (ch)
		{
			case 'A' : mode=mode+1; if (mode>3) mode=1; break;
			case 'B' : ColorEdit(); break;
			case '1' : mode=1; break;
			case '2' : mode=2; break;
			case '3' : mode=2; break;
			case 'C' : if (mode) {ClearDisplay(); while (getchar()!='C') ;} break;
		}
	}
}
